// Scilab ( http://www.scilab.org/ ) - This file is part of Scilab
// Copyright (C) DIGITEO - 2010 - Ievgen IVANOV <ivanov.eugen@gmail.com>
//
// This file must be used under the terms of the CeCILL.
// This source file is licensed as described in the file COPYING, which
// you should have received as part of this distribution.  The terms
// are also available at
// http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt


//In general we have to generate an Xcos block diagram from hybrid automaton state diagram.
//A state diagram gives state equations and jump conditions.
//A generated (main) block diagram should contain AUTOMAT block and a number of
//superblocks (AUTOMAT subsystems), each corresponing to one hybrid automaton"s state (for
//example see www.cse.iitb.ac.in/~ravikantn/Hybrid%20automata.pdf). 
//A subsystem diagram behind superblock, which corresponds to some state, encodes
//equations of this state and jump coditions from this state to other states. 
//Hence, to generate a main block diagram, first we have to generate subsystem diagrams.
//Subsystem diagram (for continuous state) looks like follows:
//                 /->--expression blocks for ODEs------->-------mux \
// in->-(splitters)-->-------------extractor--------------->--------mux->-out
//           ^     \->--expression blocks for jump conditions-->-mux /
// optional in2 (control)
//We generate it in 2 steps: 
//1) Generate an (high-level) intermediate representation of subsystem diagram
//(blockls,connls) with hybm_gen_subsys function. 
//2) Transform intermediate representation into actual diagram (scs_m structure) with
//hybm_mkdgr function
//The reason for 2-staged procedure is that scs_m structure contains a lot of low-level
//details and therefore is not convinient for generation.

//hybm_gen_subsys function generatates a representation of one AUTOMAT subsystem in the
//form of block and connection lists, which can be used as input to hybm_mkdgr.
//Parameters:
//st_eqs - column vectors of strings, right-hand sizes of ODEs (in variables
//  u1,u2,...)
//jcond - column vector of strings, jump conditions 
//  in variables u1,u2,... (state) and v1,v2,.. (control).
//insz - (optional) size of input control vector (default 0)
//To view a diagram, generated by this function, run
//loadScicosLibs;exec("SCI/modules/scicos_blocks/macros/Misc/EXPRESSION.sci");
//[blockls,connls]=hybm_gen_subsys(["-u1";"u2"],["u1+u2"]);xcos(hybm_mkdgr(blockls,connls))
function [blockls,connls]=hybm_gen_subsys(st_eqs, jconds, insz)
if argn(2)<>2 & argn(2)<>3 then
	error(msprintf(gettext("%s: Wrong number of input arguments: %d expected.\n"),"hybm_gen_subsys",2))
end
if argn(2)<3 then
	insz=0
end
if type(st_eqs)<>10 then
	error(msprintf(gettext("%s: Wrong type for input argument #%d: String array expected.\n"),"hybm_gen_subsys",1));
end
if size(st_eqs(1,:),"*")<>1 then
	error(msprintf(gettext("%s: Wrong size for input argument #%d: Non-empty column of strings expected.\n"),"hybm_gen_subsys",1));
end
if jconds<>[] & type(jconds)<>10 then
	error(msprintf(gettext("%s: Wrong type for input argument #%d: String array expected.\n"),"hybm_gen_subsys",2));
end
if size(jconds(1,:),"*")>1 then
	error(msprintf(gettext("%s: Wrong size for input argument #%d: Column of strings expected.\n"),"hybm_gen_subsys",2));
end
if type(insz)<>1 then
	error(msprintf(gettext("%s: Wrong type for input argument #%d: Number expected.\n"),"hybm_gen_subsys",3));
end
if size(insz,"*")<>1 then
	error(msprintf(gettext("%s: Wrong size for input argument #%d: Number expected.\n"),"hybm_gen_subsys",3));
end
insz=int(insz)
m=size(st_eqs,1)
m1=size(jconds,1)
vars=""; //variable names, used in state_eqs (u1,u2,..) - required for expression blocks
for i=1:2*m+insz do //last insz variable names are used for input control
	if i>1 then vars=vars+",", end
	vars=vars+"u"+sci2exp(i)
end
//add "in" block and first splitter; syntax:
//list("<blockname>",[x,y],<blocktype>,<block_parameters>) - for details see hybm_mkdgr
//Hard-coded numbers in the following expressions (10,150,etc.) are (x,y) coordinate
//values for automatic placement of blocks on a diagram.
//They were taken from manually created Xcos diagrams.
if insz>0 //merge state (#1) and control (#2) inputs (if we have control input)
	blockls=list(list("in1",[10,60*(m1+1)],"In",1),list("in2",[10,60*(m1)],"In",2))
	blockls($+1)=list("in",[70,60*(m1+1)],"Mux",2)
	connls=list(list("in1",1,"in",1),list("in2",1,"in",2))
else //use one input if there is not control input
	blockls=list(list("in",[10,60*(m1+1)],"In",1))
	connls=list()
end
blockls($+1)=list("splx",[120,60*(m1+1)],"Split")
for j=1:m1 do
	//add splitters for each jump condition expression
	blockls($+1)=list("splj"+sci2exp(j),[120,60*(m1+1)],"Split")
end
for j=1:(m-1) do
	//add splitters for expressions, which represent ODEs
	blockls($+1)=list("sple"+sci2exp(j),[120,60*(m1+1)],"Split")
end
for j=1:m do
	s="-u"+sci2exp(j+m)+"+("+st_eqs(j)+")"
	//add expression blocks for ODEs
	blockls($+1)=list("ex"+sci2exp(j),[200,60*(m-j+2+m1)],"Expr",vars,hybm_prepexpr(s,2*m,insz))
end
blockls($+1)=list("extr",[200,60*(m1+1)],"Extr",[1:m]) //add extractor
for j=1:m1 do
	//add jump conditions
	blockls($+1)=list("jmp"+sci2exp(j),[200,60*(m1-j+1)],"Expr",vars,hybm_prepexpr(jconds(j),2*m,insz))
end
//add connections in->splx->extr; syntax: list("src_name",src_port,"dest_name",dest_port)
connls=lstcat(connls,list(list("in",1,"splx",1),list("splx",1,"extr",1)))
for j=1:m1 do //connect splitters for jump condtion expressions
	if j>1 then
		connls($+1)=list("splj"+sci2exp(j-1),1,"splj"+sci2exp(j),1)
	else
		connls($+1)=list("splx",2,"splj1",1)
	end
	connls($+1)=list("splj"+sci2exp(j),2,"jmp"+sci2exp(j),1)
end
for j=1:(m-1) do //connect splitters for ODE expressions
	if j>1 then
		connls($+1)=list("sple"+sci2exp(j-1),1,"sple"+sci2exp(j),1)
	else
		if m1>0 then
			connls($+1)=list("splj"+sci2exp(m1),1,"sple1",1)
		else
			connls($+1)=list("splx",2,"sple1",1)
		end
	end
	connls($+1)=list("sple"+sci2exp(j),2,"ex"+sci2exp(j),1)
end
if m>1 then
	connls($+1)=list("sple"+sci2exp(m-1),1,"ex"+sci2exp(m),1)
else //we can omit one splitter if there is only one (1D) ODE
	if m1>0 then
		connls($+1)=list("splj"+sci2exp(m1),1,"ex1",1)
	else
		connls($+1)=list("splx",2,"ex1",1)
	end
end
if m1>1 then
	blockls($+1)=list("muxj",[320,100],"Mux",m1)
	for j=1:m1 do
		connls($+1)=list("jmp"+sci2exp(j),1,"muxj",j)
	end
	connls($+1)=list("muxj",1,"muxo",3)
elseif m1>0 then //we can omit one multiplexor if there is only one jump condition
	connls($+1)=list("jmp1",1,"muxo",3)
end
if m>1 then
	blockls($+1)=list("muxe",[320,200],"Mux",m)
	for j=1:m do
		connls($+1)=list("ex"+sci2exp(j),1,"muxe",j)
	end
	connls($+1)=list("muxe",1,"muxo",1)
else //we can omit one more multiplexor if there is only one (1D) ODE
	connls($+1)=list("ex1",1,"muxo",1)
end
if m1>0 then
	blockls($+1)=list("muxo",[400,150],"Mux",3)
else
	blockls($+1)=list("muxo",[400,150],"Mux",2)
end
connls($+1)=list("extr",1,"muxo",2) //extractor->final multiplexor(muxo)
blockls($+1)=list("out",[500,150],"Out",1)
connls($+1)=list("muxo",1,"out",1) //final multiplexor->output
endfunction
